/* Inaurabis - Source
 * Copyright (C) 2011 Daniel Collins <solemnwarning@solemnwarning.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <windows.h>
#include <commctrl.h>
#include <windowsx.h>
#include <il/il.h>
#include <il/ilu.h>
#include <il/ilut.h>

#include "resource.h"

#define DEFAULT_OPACITY 120

#define ID_BTN_OPEN 1
#define ID_BTN_SHOWHIDE 2
#define ID_BTN_MOVE 3
#define ID_BTN_SCALE 4
#define ID_BTN_OPTIONS 5
#define ID_BTN_ABOUT 6
#define ID_OPAC_EDIT 7
#define ID_OPAC_UPDOWN 8
#define ID_TOOLTIP 11

/* DevIL image IDs */
#define ORIGINAL_IMAGE 1
#define SCALED_IMAGE 2

/* Toolbar layouts */
#define LAYOUT_VERTICAL 0
#define LAYOUT_NARROW 1
#define LAYOUT_HORIZONTAL 2

#define ALPHA_SKIP_EDIT (int)(1 << 0)

HWND toolbar_window = NULL;
HWND overlay_window = NULL;
HWND overlay_static = NULL;

bool image_loaded = false;
bool overlay_visible = false;
bool dialog_open = false;

HKEY regkey = NULL;

DWORD opt_toolbar_layout;
DWORD opt_toolbar_topmost;
DWORD opt_restore_state;

INT_PTR CALLBACK resize_dproc(HWND window, UINT msg, WPARAM wp, LPARAM lp);
INT_PTR CALLBACK options_dproc(HWND window, UINT msg, WPARAM wp, LPARAM lp);
INT_PTR CALLBACK about_dproc(HWND window, UINT msg, WPARAM wp, LPARAM lp);

void register_classes(HINSTANCE instance);
bool scale_image(int width, int height);
bool load_image(const char *filename, int w = 0, int h = 0);
void set_alpha(int alpha, int flags = 0);
void enable_buttons();
int edit_get_int(HWND window, HWND control, int min_val, int max_val, bool silent_fail);
void edit_set_int(HWND control, int val);
void error_msgbox(const char *title, const char *fmt, ...);
const char *win32_errmsg(DWORD errnum = GetLastError());

DWORD reg_get_dword(const char *name, DWORD default_val);
bool reg_set_dword(const char *name, DWORD value);
char *reg_get_string(const char *name, const char *default_val);
bool reg_set_string(const char *name, const char *value);
void reg_create();

#define error_dialog(...) error_msgbox("Error", __VA_ARGS__)
#define fatal_error(...) { error_msgbox("Fatal error", __VA_ARGS__); exit(1); }

void set_tooltip(HWND window, HWND control, const char *text) {
	HWND tooltip = GetDlgItem(window, ID_TOOLTIP);
	if(!tooltip) {
		tooltip = CreateWindowEx(
			WS_EX_TOPMOST,
			TOOLTIPS_CLASS,
			NULL,
			WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP,
			CW_USEDEFAULT, CW_USEDEFAULT,
			CW_USEDEFAULT, CW_USEDEFAULT,
			window,
			NULL,
			GetModuleHandle(NULL),
			NULL
		);
		
		if(!tooltip) {
			return;
		}
	}
	
	TOOLINFO ti;
	
	memset(&ti, 0, sizeof(ti));
	ti.cbSize = sizeof(ti);
	ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
	ti.hwnd = window;
	ti.uId = (UINT_PTR)control;
	ti.lpszText = (char*)text;
	
	SendMessage(tooltip, TTM_ADDTOOL, 0, (LPARAM)&ti);
}

void icon_button(HWND parent, int id, int x, int y, const char *text, int icon_id) {
	HWND button = GetDlgItem(parent, id);
	
	if(button) {
		SetWindowPos(button, 0, x, y, 0, 0, SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
	}else{
		button = CreateWindow("BUTTON", text, WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | BS_ICON, x, y, 28, 28, parent, (HMENU)id, NULL, NULL);
		
		HANDLE icon = (HBITMAP)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(icon_id), IMAGE_ICON, 0, 0, LR_SHARED | LR_LOADMAP3DCOLORS);
		SendMessage(button, BM_SETIMAGE, IMAGE_ICON, (LRESULT)icon);
		
		set_tooltip(parent, button, text);
	}
}

#define BTN_OFFSET(n) (2 + n * (btn_size + 4))

void init_toolbar(HWND window, int layout) {
	int edge = GetSystemMetrics(SM_CYEDGE);
	int btn_size = 24 + 2 * edge;
	
	int edit_w = btn_size * 2 + 4;
	int edit_h = btn_size;
	
	int window_w, window_h;
	
	HWND alpha_edit = GetDlgItem(window, ID_OPAC_EDIT);
	if(!alpha_edit) {
		alpha_edit = CreateWindowEx(WS_EX_CLIENTEDGE, "EDIT", "", WS_CHILD | WS_VISIBLE | WS_BORDER | ES_NUMBER, 0, 0, edit_w, edit_h, window, (HMENU)ID_OPAC_EDIT, NULL, NULL);
		
		SendMessage(alpha_edit, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT), TRUE);
		
		set_tooltip(window, alpha_edit, "Overlay opacity (0-255)");
	}
	
	HWND updown = GetDlgItem(window, ID_OPAC_UPDOWN);
	if(!updown) {
		updown = CreateWindowEx(
			WS_EX_LEFT,
			UPDOWN_CLASS,
			NULL,
			WS_CHILD | WS_VISIBLE | WS_BORDER | UDS_ALIGNRIGHT | UDS_ARROWKEYS | UDS_HOTTRACK,
			0, 0,
			0, 0,
			window,
			(HMENU)ID_OPAC_UPDOWN,
			GetModuleHandle(NULL),
			NULL
		);
		
		SendMessage(updown, UDM_SETRANGE, 0, MAKELPARAM(255, 0));
		
		set_tooltip(window, updown, "Increase/decrease overlay opacity");
	}
	
	switch(layout) {
		case LAYOUT_VERTICAL:
			icon_button(window, ID_BTN_OPEN, 2, 2, "Open image", ICON_OPEN);
			icon_button(window, ID_BTN_SHOWHIDE, BTN_OFFSET(1), 2, "Show/Hide overlay", ICON_SHOWHIDE);
			icon_button(window, ID_BTN_MOVE, 2, BTN_OFFSET(1), "Move overlay", ICON_MOVE);
			icon_button(window, ID_BTN_SCALE, BTN_OFFSET(1), BTN_OFFSET(1), "Resize overlay", ICON_RESIZE);
			icon_button(window, ID_BTN_OPTIONS, 2, BTN_OFFSET(2), "Options", ICON_OPTIONS);
			icon_button(window, ID_BTN_ABOUT, BTN_OFFSET(1), BTN_OFFSET(2), "About", ICON_ABOUT);
			
			SetWindowPos(alpha_edit, 0, 2, BTN_OFFSET(3), edit_w, edit_h, SWP_NOZORDER | SWP_NOACTIVATE);
			
			window_w = 2 * (btn_size + 4);
			window_h = 4 * (btn_size + 4);
			
			break;
			
		case LAYOUT_NARROW:
			icon_button(window, ID_BTN_OPEN, 2, 2, "Open image", ICON_OPEN);
			icon_button(window, ID_BTN_SHOWHIDE, 2, BTN_OFFSET(1), "Show/Hide overlay", ICON_SHOWHIDE);
			icon_button(window, ID_BTN_MOVE, 2, BTN_OFFSET(2), "Move overlay", ICON_MOVE);
			icon_button(window, ID_BTN_SCALE, 2, BTN_OFFSET(3), "Resize overlay", ICON_RESIZE);
			icon_button(window, ID_BTN_OPTIONS, 2, BTN_OFFSET(4), "Options", ICON_OPTIONS);
			icon_button(window, ID_BTN_ABOUT, 2, BTN_OFFSET(5), "About", ICON_ABOUT);
			
			window_w = 1 * (btn_size + 4);
			window_h = 7 * (btn_size + 4);
			
			break;
			
		case LAYOUT_HORIZONTAL:
			icon_button(window, ID_BTN_OPEN, 2, 2, "Open image", ICON_OPEN);
			icon_button(window, ID_BTN_SHOWHIDE, BTN_OFFSET(1), 2, "Show/Hide overlay", ICON_SHOWHIDE);
			icon_button(window, ID_BTN_MOVE, BTN_OFFSET(2), 2, "Move overlay", ICON_MOVE);
			icon_button(window, ID_BTN_SCALE, BTN_OFFSET(3), 2, "Resize overlay", ICON_RESIZE);
			icon_button(window, ID_BTN_OPTIONS, BTN_OFFSET(4), 2, "Options", ICON_OPTIONS);
			icon_button(window, ID_BTN_ABOUT, BTN_OFFSET(5), 2, "About", ICON_ABOUT);
			
			SetWindowPos(alpha_edit, 0, BTN_OFFSET(6), 2, edit_w, edit_h, SWP_NOZORDER | SWP_NOACTIVATE);
			
			window_w = 8 * (btn_size + 4);
			window_h = 1 * (btn_size + 4);
			
			break;
			
		default:
			break;
	}
	
	if(layout == LAYOUT_NARROW) {
		ShowWindow(alpha_edit, SW_HIDE);
		SendMessage(updown, UDM_SETBUDDY, (WPARAM)NULL, 0);
		
		SetWindowPos(updown, 0, 2, BTN_OFFSET(6), btn_size, btn_size, SWP_NOZORDER | SWP_NOACTIVATE);
	}else{
		ShowWindow(alpha_edit, SW_SHOWNA);
		SendMessage(updown, UDM_SETBUDDY, (WPARAM)alpha_edit, 0);
		
		/* Hack to make updown control the correct height, is there a better way? */
		
		RECT rect;
		
		GetWindowRect(updown, &rect);
		MapWindowPoints(HWND_DESKTOP, window, (POINT*)&rect, 2);
		
		int ud_w = rect.right - rect.left;
		int ud_h = (rect.bottom - rect.top) - 2;
		
		SetWindowPos(updown, 0, rect.left, rect.top + 1, ud_w, ud_h, SWP_NOZORDER | SWP_NOACTIVATE);
	}
	
	/* Resize toolbar window, account for borders */
	
	RECT rect, c_rect;
	
	GetWindowRect(window, &rect);
	GetClientRect(window, &c_rect);
	
	window_w += (rect.right - rect.left) - c_rect.right;
	window_h += (rect.bottom - rect.top) - c_rect.bottom;
	
	SetWindowPos(window, 0, 0, 0, window_w, window_h, SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
}

LRESULT CALLBACK cmd_wproc(HWND window, UINT msg, WPARAM wp, LPARAM lp) {
	switch(msg) {
		case WM_CREATE: {
			init_toolbar(window, opt_toolbar_layout);
			break;
		}
		
		case WM_COMMAND: {
			if(HIWORD(wp) == BN_CLICKED) {
				switch(LOWORD(wp)) {
					case ID_BTN_OPEN: {
						char filename[512] = "";
						
						OPENFILENAME openfile;
						
						memset(&openfile, 0, sizeof(openfile));
						openfile.lStructSize = sizeof(openfile);
						openfile.hwndOwner = window;
						openfile.lpstrFilter = "All Files\0*\0";
						openfile.lpstrFile = filename;
						openfile.nMaxFile = sizeof(filename);
						openfile.lpstrTitle = "Load image";
						openfile.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
						
						if(GetOpenFileName(&openfile)) {
							load_image(filename);
						}else if(CommDlgExtendedError()) {
							error_dialog("GetOpenFileName: %u", (unsigned int)CommDlgExtendedError());
						}
						
						break;
					}
					
					case ID_BTN_SHOWHIDE: {
						overlay_visible = !overlay_visible;
						ShowWindow(overlay_window, overlay_visible ? SW_SHOWNA : SW_HIDE);
						
						reg_set_dword("last_show", overlay_visible);
						
						enable_buttons();
						
						break;
					}
					
					case ID_BTN_MOVE: {
						SetWindowLongPtr(overlay_window, GWL_EXSTYLE, WS_EX_LAYERED | WS_EX_TOPMOST);
						break;
					}
					
					case ID_BTN_SCALE: {
						dialog_open = true;
						DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(DIALOG_RESIZE), window, &resize_dproc);
						dialog_open = false;
						
						break;
					}
					
					case ID_BTN_OPTIONS: {
						dialog_open = true;
						DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(DIALOG_OPTIONS), window, &options_dproc);
						dialog_open = false;
						
						break;
					}
					
					case ID_BTN_ABOUT: {
						dialog_open = true;
						DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(DIALOG_ABOUT), window, &about_dproc);
						dialog_open = false;
						
						break;
					}	
					
					default:
						break;
				}
			}else if(HIWORD(wp) == EN_CHANGE && LOWORD(wp) == ID_OPAC_EDIT) {
				int alpha = edit_get_int(window, (HWND)lp, 0, 255, true);
				if(alpha != -1) {
					set_alpha(alpha, ALPHA_SKIP_EDIT);
				}
			}
			
			break;
		}
		
		case WM_NOTIFY: {
			NMHDR *notify = (NMHDR*)lp;
			
			if(notify->idFrom == ID_OPAC_UPDOWN && notify->code == UDN_DELTAPOS) {
				NMUPDOWN *updown = (NMUPDOWN*)lp;
				int alpha = updown->iPos + updown->iDelta;
				
				if(alpha > 255) {
					alpha = 255;
				}else if(alpha < 0) {
					alpha = 0;
				}
				
				set_alpha(alpha);
				return TRUE;
				
			}
			
			break;
		}
		
		case WM_SETFOCUS: {
			SetWindowPos(window, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
			break;
		}
		
		case WM_KILLFOCUS: {
			if(!opt_toolbar_topmost && !dialog_open) {
				SetWindowPos(window, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
			}
			
			break;
		}
		
		case WM_CLOSE: {
			DestroyWindow(window);
			break;
		}
		
		case WM_DESTROY: {
			PostQuitMessage(0);
			break;
		}
		
		default:
			return DefWindowProc(window, msg, wp, lp);
	};
	
	return 0;
}

LRESULT CALLBACK overlay_wproc(HWND window, UINT msg, WPARAM wp, LPARAM lp) {
	static bool moving = false;
	static int init_x, init_y;
	
	switch(msg) {
		case WM_CLOSE:
			DestroyWindow(window);
			break;
		
		case WM_LBUTTONUP:
		case WM_KILLFOCUS:
			moving = false;
			SetWindowLongPtr(window, GWL_EXSTYLE, WS_EX_TRANSPARENT | WS_EX_LAYERED | WS_EX_TOPMOST);
			
			if(opt_toolbar_topmost) {
				SetWindowPos(toolbar_window, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
			}
			
			RECT cur_pos;
			GetWindowRect(window, &cur_pos);
			
			reg_set_dword("last_x", cur_pos.left);
			reg_set_dword("last_y", cur_pos.top);
			
			break;
		
		case WM_LBUTTONDOWN:
			moving = true;
			init_x = LOWORD(lp);
			init_y = HIWORD(lp);
			
			break;
		
		case WM_MOUSEMOVE: {
			if(moving) {
				int mov_x = LOWORD(lp) - init_x;
				int mov_y = HIWORD(lp) - init_y;
				
				RECT cur_pos;
				GetWindowRect(window, &cur_pos);
				SetWindowPos(window, HWND_TOPMOST, cur_pos.left + mov_x, cur_pos.top + mov_y, 0, 0, SWP_NOSIZE);
			}
			
			break;
		}
		
		case WM_SETFOCUS: {
			if(opt_toolbar_topmost) {
				SetWindowPos(toolbar_window, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
			}
			
			break;
		}
		
		default:
			return DefWindowProc(window, msg, wp, lp);
	};
	
	return 0;
}

INT_PTR CALLBACK resize_dproc(HWND window, UINT msg, WPARAM wp, LPARAM lp) {
	static int init_w, init_h, last_w, last_h;
	
	switch(msg) {
		case WM_INITDIALOG: {
			RECT rect;
			GetWindowRect(overlay_static, &rect);
			
			last_w = init_w = rect.right - rect.left;
			last_h = init_h = rect.bottom - rect.top;
			
			edit_set_int(GetDlgItem(window, EDIT_WIDTH), init_w);
			edit_set_int(GetDlgItem(window, EDIT_HEIGHT), init_h);
			
			Button_SetCheck(GetDlgItem(window, OPT_ASPECT), BST_CHECKED);
			
			break;
		}
		
		case WM_COMMAND: {
			if(HIWORD(wp) == BN_CLICKED) {
				switch(LOWORD(wp)) {
					case BTN_OK: {
						int w, h;
						
						if((w = edit_get_int(window, GetDlgItem(window, EDIT_WIDTH), 1, INT_MAX, false)) == -1 || (h = edit_get_int(window, GetDlgItem(window, EDIT_HEIGHT), 1, INT_MAX, false)) == -1) {
							break;
						}
						
						if(w != init_w || h != init_h) {
							scale_image(w, h);
						}
					}
					
					case BTN_CANCEL: {
						EndDialog(window, 0);
						return TRUE;
					}
					
					default:
						break;
				}
			}else if(HIWORD(wp) == EN_KILLFOCUS) {
				/* Enforce aspect ratio */
				
				if(!Button_GetCheck(GetDlgItem(window, OPT_ASPECT))) {
					break;
				}
				
				double w, h;
				
				if(LOWORD(wp) == EDIT_WIDTH) {
					if((w = edit_get_int(window, GetDlgItem(window, EDIT_WIDTH), 1, INT_MAX, true)) == -1 || w == last_w) {
						break;
					}
					
					h = init_h;
					h /= init_w;
					h *= w;
					
					last_w = w;
					last_h = h;
					
					edit_set_int(GetDlgItem(window, EDIT_HEIGHT), h);
				}else if(LOWORD(wp) == EDIT_HEIGHT) {
					if((h = edit_get_int(window, GetDlgItem(window, EDIT_HEIGHT), 1, INT_MAX, true)) == -1 || h == last_h) {
						break;
					}
					
					w = init_w;
					w /= init_h;
					w *= h;
					
					last_w = w;
					last_h = h;
					
					edit_set_int(GetDlgItem(window, EDIT_WIDTH), w);
				}
			}
			
			break;
		}
		
		default:
			break;
	}
	
	return FALSE;
}

INT_PTR CALLBACK options_dproc(HWND window, UINT msg, WPARAM wp, LPARAM lp) {
	switch(msg) {
		case WM_INITDIALOG: {
			switch(opt_toolbar_layout) {
				case LAYOUT_VERTICAL:
					Button_SetCheck(GetDlgItem(window, LAYOUT_BTN_VERT), BST_CHECKED);
					break;
					
				case LAYOUT_NARROW:
					Button_SetCheck(GetDlgItem(window, LAYOUT_BTN_NARROW), BST_CHECKED);
					break;
					
				case LAYOUT_HORIZONTAL:
					Button_SetCheck(GetDlgItem(window, LAYOUT_BTN_HORIZ), BST_CHECKED);
					break;
					
				default:
					break;
			};
			
			Button_SetCheck(GetDlgItem(window, OPT_TOPMOST), opt_toolbar_topmost);
			Button_SetCheck(GetDlgItem(window, OPT_REMEMBER), opt_restore_state);
			
			break;
		}
		
		case WM_COMMAND: {
			if(HIWORD(wp) == BN_CLICKED) {
				switch(LOWORD(wp)) {
					case BTN_OK: {
						if(!regkey) {
							reg_create();
						}
						
						if(Button_GetCheck(GetDlgItem(window, LAYOUT_BTN_VERT))) {
							opt_toolbar_layout = LAYOUT_VERTICAL;
						}else if(Button_GetCheck(GetDlgItem(window, LAYOUT_BTN_NARROW))) {
							opt_toolbar_layout = LAYOUT_NARROW;
						}else{
							opt_toolbar_layout = LAYOUT_HORIZONTAL;
						}
						
						opt_toolbar_topmost = Button_GetCheck(GetDlgItem(window, OPT_TOPMOST));
						opt_restore_state = Button_GetCheck(GetDlgItem(window, OPT_REMEMBER));
						
						init_toolbar(toolbar_window, opt_toolbar_layout);
						
						reg_set_dword("toolbar_layout", opt_toolbar_layout);
						reg_set_dword("toolbar_topmost", opt_toolbar_topmost);
						reg_set_dword("restore_state", opt_restore_state);
					}
					
					case BTN_CANCEL: {
						EndDialog(window, 0);
						return TRUE;
					}
					
					default:
						break;
				}
			}
			
			break;
		}
		
		default:
			break;
	};
	
	return FALSE;
}

INT_PTR CALLBACK about_dproc(HWND window, UINT msg, WPARAM wp, LPARAM lp) {
	switch(msg) {
		case WM_INITDIALOG: {
			/* TODO: Set version */
			break;
		}
		
		case WM_COMMAND: {
			if(HIWORD(wp) == BN_CLICKED && LOWORD(wp) == BTN_CLOSE) {
				EndDialog(window, 0);
				return TRUE;
			}
			
			break;
		}
		
		default:
			break;
	};
	
	return FALSE;
}

void register_classes(HINSTANCE instance) {
	INITCOMMONCONTROLSEX cc;
	cc.dwSize = sizeof(cc);
	cc.dwICC = ICC_BAR_CLASSES | ICC_UPDOWN_CLASS;
	
	if(!InitCommonControlsEx(&cc)) {
		fatal_error("Failed to initialise common controls");
	}
	
	WNDCLASSEX cmd_style, overlay_style;
	
	memset(&cmd_style, 0, sizeof(cmd_style));
	cmd_style.cbSize = sizeof(cmd_style);
	cmd_style.lpfnWndProc = &cmd_wproc;
	cmd_style.hInstance = instance;
	cmd_style.hIcon = (HICON)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(ICON_INAURABIS32), IMAGE_ICON, 0, 0, LR_SHARED);
	cmd_style.hCursor = LoadCursor(NULL, IDC_ARROW);
	cmd_style.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
	cmd_style.lpszMenuName = NULL;
	cmd_style.lpszClassName = "inaurabis_toolbar_class";
	cmd_style.hIconSm = (HICON)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(ICON_INAURABIS16), IMAGE_ICON, 0, 0, LR_SHARED);
	
	if(!RegisterClassEx(&cmd_style)) {
		fatal_error("Failed to register inaurabis_toolbar_class: %s", win32_errmsg());
	}
	
	memset(&overlay_style, 0, sizeof(overlay_style));
	overlay_style.cbSize = sizeof(overlay_style);
	overlay_style.lpfnWndProc = &overlay_wproc;
	overlay_style.hInstance = instance;
	overlay_style.hIcon = NULL;
	overlay_style.hCursor = LoadCursor(NULL, IDC_SIZEALL);
	overlay_style.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
	overlay_style.lpszMenuName = NULL;
	overlay_style.lpszClassName = "inaurabis_overlay_class";
	overlay_style.hIconSm = NULL;
	
	if(!RegisterClassEx(&overlay_style)) {
		fatal_error("Failed to register inaurabis_overlay_class: %s", win32_errmsg());
	}
}

bool scale_image(int width, int height) {
	ilBindImage(SCALED_IMAGE);
	
	if(!ilCopyImage(ORIGINAL_IMAGE)) {
		error_dialog("Image copying failed: %s", iluErrorString(ilGetError()));
		return false;
	}
	
	iluImageParameter(ILU_FILTER, ILU_SCALE_LANCZOS3);
	
	if(!iluScale(width, height, ilGetInteger(IL_IMAGE_DEPTH))) {
		error_dialog("Image scaling failed: %s", iluErrorString(ilGetError()));
		return false;
	}
	
	HBITMAP bitmap = ilutConvertToHBitmap(GetDC(NULL));
	if(!bitmap) {
		error_dialog("Bitmap conversion failed: %s", iluErrorString(ilGetError()));
		return false;
	}
	
	SetWindowPos(overlay_window, HWND_TOPMOST, 0, 0, width, height, SWP_NOMOVE);
	SetWindowPos(overlay_static, 0, 0, 0, width, height, SWP_NOMOVE | SWP_NOZORDER);
	
	HBITMAP old_bitmap = (HBITMAP)SendMessage(overlay_static, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)bitmap);
	if(old_bitmap) {
		DeleteObject(old_bitmap);
	}
	
	reg_set_dword("last_w", width);
	reg_set_dword("last_h", height);
	
	return true;
}

bool load_image(const char *filename, int w, int h) {
	ilBindImage(ORIGINAL_IMAGE);
	
	if(!ilLoadImage(filename)) {
		error_dialog("Image loading failed: %s", iluErrorString(ilGetError()));
		return false;
	}
	
	if(!scale_image(w ? w : ilGetInteger(IL_IMAGE_WIDTH), h ? h : ilGetInteger(IL_IMAGE_HEIGHT))) {
		return false;
	}
	
	reg_set_string("last_image", filename);
	
	image_loaded = true;
	enable_buttons();
	
	return true;
}

void set_alpha(int alpha, int flags) {
	SetLayeredWindowAttributes(overlay_window, RGB(255,255,255), alpha, LWA_ALPHA);
	
	SetWindowPos(overlay_window, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	SetWindowPos(toolbar_window, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	
	SendMessage(GetDlgItem(toolbar_window, ID_OPAC_UPDOWN), UDM_SETPOS, 0, alpha);
	
	if(!(flags & ALPHA_SKIP_EDIT)) {
		char alpha_text[16];
		sprintf(alpha_text, "%d", alpha);
		
		SetWindowText(GetDlgItem(toolbar_window, ID_OPAC_EDIT), alpha_text);
	}
	
	reg_set_dword("last_alpha", alpha);
}

void enable_buttons() {
	EnableWindow(GetDlgItem(toolbar_window, ID_BTN_SHOWHIDE), image_loaded);
	EnableWindow(GetDlgItem(toolbar_window, ID_BTN_MOVE), image_loaded && overlay_visible);
	EnableWindow(GetDlgItem(toolbar_window, ID_BTN_SCALE), image_loaded);
	
	EnableWindow(GetDlgItem(toolbar_window, ID_OPAC_EDIT), image_loaded);
	EnableWindow(GetDlgItem(toolbar_window, ID_OPAC_UPDOWN), image_loaded);
}

int edit_get_int(HWND window, HWND control, int min_val, int max_val, bool silent_fail) {
	char buf[64];
	GetWindowText(control, buf, sizeof(buf));
	
	int i = atoi(buf);
	
	if(buf[0] == '\0' || buf[strspn(buf, "1234567890")] != '\0' || i < min_val || i > max_val) {
		if(!silent_fail) {
			sprintf(buf, "Input must be an integer between %d and %d", min_val, max_val);
			MessageBox(window, buf, NULL, MB_OK | MB_ICONERROR);
			
			SetFocus(control);
			Edit_SetSel(control, 0, Edit_GetTextLength(control));
		}
		
		return -1;
	}
	
	return i;
}

void edit_set_int(HWND control, int val) {
	char buf[64];
	sprintf(buf, "%d", val);
	
	SetWindowText(control, buf);
}

void error_msgbox(const char *title, const char *fmt, ...) {
	char buf[256];
	
	va_list argv;
	va_start(argv, fmt);
	vsnprintf(buf, sizeof(buf), fmt, argv);
	va_end(argv);
	
	MessageBox(NULL, buf, title, MB_OK | MB_ICONERROR | MB_TASKMODAL);
}

const char *win32_errmsg(DWORD errnum) {
	static char buf[256];
	
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, errnum, 0, buf, sizeof(buf), NULL);
	return buf;
}

DWORD reg_get_dword(const char *name, DWORD default_val) {
	if(!regkey) {
		return default_val;
	}
	
	DWORD buf, size = sizeof(buf), err;
	
	if((err = RegQueryValueEx(regkey, name, NULL, NULL, (BYTE*)&buf, &size)) != ERROR_SUCCESS) {
		if(err != ERROR_FILE_NOT_FOUND) {
			error_dialog("Error reading registry value: %s", win32_errmsg(err));
		}
		
		return default_val;
	}
	
	return buf;
}

bool reg_set_dword(const char *name, DWORD value) {
	if(!regkey) {
		return false;
	}
	
	DWORD err = RegSetValueEx(regkey, name, 0, REG_DWORD, (BYTE*)&value, sizeof(value));
	if(err != ERROR_SUCCESS) {
		error_dialog("Error writing registry value: %s", win32_errmsg(err));
		return false;
	}
	
	return true;
}

/* Read a string value from the registry
 *
 * Returns a copy of default_val in memory allocated with malloc if the value doesn't
 * exist or an error occurs. (No copying is performed if default_val is NULL)
*/
char *reg_get_string(const char *name, const char *default_val) {
	DWORD size = 0, err;
	char *buf;
	
	if(!regkey) {
		goto RETURN_DEFAULT;
	}
	
	if((err = RegQueryValueEx(regkey, name, NULL, NULL, NULL, &size)) != ERROR_SUCCESS && err != ERROR_MORE_DATA) {
		if(err != ERROR_FILE_NOT_FOUND) {
			error_dialog("Error reading registry value: %s", win32_errmsg(err));
		}
		
		goto RETURN_DEFAULT;
	}
	
	buf = (char*)malloc(size+1);
	buf[size] = '\0';
	
	if((err = RegQueryValueEx(regkey, name, NULL, NULL, (BYTE*)buf, &size)) != ERROR_SUCCESS) {
		free(buf);
		goto RETURN_DEFAULT;
	}
	
	return buf;
	
	RETURN_DEFAULT:
	
	if(default_val) {
		char *ret = (char*)malloc(strlen(default_val) + 1);
		strcpy(ret, default_val);
		
		return ret;
	}
	
	return NULL;
}

bool reg_set_string(const char *name, const char *value) {
	if(!regkey) {
		return false;
	}
	
	DWORD err = RegSetValueEx(regkey, name, 0, REG_SZ, (BYTE*)value, strlen(value) + 1);
	if(err != ERROR_SUCCESS) {
		error_dialog("Error writing registry value: %s", win32_errmsg(err));
		return false;
	}
	
	return true;
}

void reg_create() {
	DWORD errnum = RegCreateKeyEx(
		HKEY_CURRENT_USER,
		"SOFTWARE\\Inaurabis",
		0,
		NULL,
		0,
		KEY_QUERY_VALUE | KEY_SET_VALUE,
		NULL,
		&regkey,
		NULL
	);
	
	if(errnum != ERROR_SUCCESS) {
		error_dialog("Error creating/opening registry key: %s", win32_errmsg(errnum));
		regkey = NULL;
	}
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow) {
	ilInit();
	iluInit();
	
	if(!ilutRenderer(ILUT_WIN32)) {
		fatal_error("Failed to initialise ILUT (DevIL): %s", iluErrorString(ilGetError()));
	}
	
	register_classes(hInstance);
	
	reg_create();
	
	opt_toolbar_layout = reg_get_dword("toolbar_layout", LAYOUT_HORIZONTAL);
	opt_toolbar_topmost = reg_get_dword("toolbar_topmost", 0);
	opt_restore_state = reg_get_dword("restore_state", 0);
	
	HWND container = CreateWindowEx(
		WS_EX_TRANSPARENT | WS_EX_LAYERED | WS_EX_TOPMOST,
		"inaurabis_overlay_class",
		"Inaurabis overlay",
		WS_POPUP,
		CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, CW_USEDEFAULT,
		NULL,
		NULL,
		hInstance,
		NULL
	);
	
	if(!container) {
		fatal_error("Failed to create container window: %s", win32_errmsg());
	}
	
	overlay_window = CreateWindowEx(
		WS_EX_TRANSPARENT | WS_EX_LAYERED | WS_EX_TOPMOST,
		"inaurabis_overlay_class",
		"Inaurabis overlay",
		WS_POPUP,
		0, 0,
		0, 0,
		container,
		NULL,
		hInstance,
		NULL
	);
	
	if(!overlay_window) {
		fatal_error("Failed to create overlay_window: %s", win32_errmsg());
	}
	
	overlay_static = CreateWindow(
		"STATIC",
		NULL,
		WS_VISIBLE | WS_CHILD | SS_BITMAP,
		0, 0,
		0, 0,
		overlay_window,
		NULL,
		hInstance,
		NULL
	);
	
	if(!overlay_static) {
		fatal_error("Failed to create overlay_static: %s", win32_errmsg());
	}
	
	toolbar_window = CreateWindowEx(
		WS_EX_TOOLWINDOW | WS_EX_APPWINDOW,
		"inaurabis_toolbar_class",
		"Inaurabis toolbar",
		WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
		CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, CW_USEDEFAULT,
		NULL,
		NULL,
		hInstance,
		NULL
	);
	
	if(!toolbar_window) {
		fatal_error("Failed to create toolbar window: %s", win32_errmsg());
	}
	
	DWORD alpha = reg_get_dword("last_alpha", 256);
	set_alpha(alpha < 256 ? alpha : DEFAULT_OPACITY);
	
	if(opt_restore_state) {
		/* Load previous image and state */
		
		char *filename = reg_get_string("last_image", NULL);
		
		if(filename) {
			DWORD width = reg_get_dword("last_w", 0);
			DWORD height = reg_get_dword("last_h", 0);
			
			load_image(filename, width, height);
			
			DWORD x = reg_get_dword("last_x", 0);
			DWORD y = reg_get_dword("last_y", 0);
			
			SetWindowPos(overlay_window, 0, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
			
			if(reg_get_dword("last_show", 0)) {
				overlay_visible = true;
				ShowWindow(overlay_window, SW_SHOWNA);
			}
			
			free(filename);
		}
	}
	
	if(opt_toolbar_topmost) {
		SetWindowPos(toolbar_window, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	}
	
	enable_buttons();
	
	ShowWindow(toolbar_window, iCmdShow);
	
	MSG msg;
	BOOL mret;
	
	while((mret = GetMessage(&msg, NULL, 0, 0))) {
		if(mret == -1) {
			fatal_error("GetMessage: %s", win32_errmsg());
		}
		
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	if(regkey) {
		RegCloseKey(regkey);
	}
	
	return 0;
}
